// SLIST.CPP
// Code for SLIST.HPP

#include "slist.hpp"


int snode::count=0;

snode::~snode()
{
  count--;
}


int length(snode *list)
{
  int count=0;
  while (list) {
    count++;
    list = list->next;
  }
  return count;
}


snode *last(snode *list)
{
  while (list && list->next)
    list = list->next;
  return list;
}


snode *nth(snode *list, int which)
{
  while (list && which--)
    list = list->next;
  return list;
}


// uses the fast-slow algorithm for finding loops
int listok(snode *list)
{
  if (!list)
    return 1;

  snode *slow = list;
  snode *fast = list->next;

  while (fast && fast->next && (fast != slow)) {
    fast = fast->next->next;
    slow = slow->next;
  }

  if (fast == slow)
    return 0;         // loop exists
  else
    return 1;         // no loops
}


snode *firstThat(snode *list, snode_predicate func, void *extra)
{
  while (list && !(*func)(list, extra))
    list = list->next;
  return list;
}


// code is, of course, very similar to firstThat but keeping
// them separate allows better self-commenting
void forAll(snode *list, snode_predicate func, void *extra)
{
  while (list && (*func)(list, extra))
    list = list->next;
}


snode *map(snode *list, snode_map func, void *extra)
{
  snode *result;
#pragma warn -pia
  while (list && !(result = (*func)(list, extra)))
    list = list->next;
#pragma warn +pia
  if (!list)
    return list;

  snode *head, *tail;
  tail = head = result;
  list = list->next;

  while (list) {
    result = (*func)(list, extra);

    if (result)
      tail = tail->next = result;

    list = list->next;
  }

  return head;
}


snode *prepend(snode *list, snode *toPrepend)
{
  toPrepend->next = list;
  return toPrepend;
}


snode *append(snode *list, snode *toAppend)
{
  snode *n = last(list);
  if (n) {
    n->next = toAppend;
    return list;
  }
  else
    return toAppend;
}


snode *remove(snode *list, snode *toRemove)
{
  if (!list)
    return list;

  if (list == toRemove) {
    snode *retval = list->next;
    toRemove->next = 0;
    return retval;
  }

  snode *n = list;
  while (n->next && (n->next != toRemove))
    n = n->next;
  if (!n->next)
    return n->next;

  n->next = n->next->next;
  toRemove->next = 0;
  return list;
}


snode *concat(snode *list1, snode *list2)
{
  if (!list1)
    return list2;

  prepend(list2, last(list1));
  return list1;
}


// this is a fairly efficient way to remove nodes; the most straightforward
// way, testing each and calling remove(), is worst-case O(n)
snode *deleteThat(snode *list, snode_predicate pred, void *extra)
{
  while (list && pred(list, extra)) {
    snode *temp = list;       // save ptr
    list = list->next;        // skip
    delete temp;              // kill
  }

  snode *p = list;
  while (p && p->next) {
    if (pred(p->next, extra)) {
      snode *temp = p->next;            // save ptr
      p->next = p->next->next;          // patch around
      delete temp;                      // kill
    }
    else
      p = p->next;                      // advance
  }

  return list;
}


// string together the nodes that pass the test and return this list also
// this function effectively partitions the list
snode *removeThat(snode *list, snode **rnodes,
                  snode_predicate pred, void *extra)
{
  snode dummy;                  // create a dummy node to simplify logic
  snode *tail = &dummy;         // tail of list of removed nodes
  dummy.next = 0;               // make sure an empty list handled properly

  while (list && pred(list, extra)) {
    snode *temp = list;              // save ptr
    list = list->next;               // skip
    tail = tail->next = temp;        // stuff in aux list
  }

  snode *p = list;
  while (p && p->next) {
    if (pred(p->next, extra)) {
      snode *temp = p->next;            // save ptr
      p->next = p->next->next;          // patch around
      tail = tail->next = temp;         // stuff in aux list
    }
    p = p->next;                     // advance
  }

  if (rnodes)
    *rnodes = dummy.next;            // dummy node goes away
  return list;
}


snode *insertat(snode *list, snode *toInsert, int destPos)
{
  if (!destPos)
    return prepend(list, toInsert);

  snode *head = list;
  while ((destPos-- > 1) && list->next)
    list = list->next;

  toInsert->next = list->next;
  list->next = toInsert;
  return head;
}


// simple-minded sorting routine
snode *insertionSort(snode *list, snode_differencer diff, void *extra)
{
  snode *primary = list;                      // primary sorting pointer
  while (primary && primary->next) {
    if (diff(primary, primary->next, extra) > 0) {   // must move next node?
      snode *tomove = primary->next;
      primary->next = primary->next->next;    // patch around moving node

      if (diff(tomove, list, extra) < 0) {           // new node goes at top?
        tomove->next = list;
        list = tomove;
      }

      else {                                  // new node *not* at top
        snode *searcher = list;
        while (diff(tomove, searcher->next, extra) > 0)
          searcher = searcher->next;

        tomove->next = searcher->next;        // insert new node into list
        searcher->next = tomove;
      }
    }

    else
      primary = primary->next;              // go on if no changes
  }

  return list;
}


static int recurse(snode *list, snode *dest,
                   snode_predicate func, void *extra)
{
  if (!list)
    return (*func)(dest, extra);

  // do first specially
  snode *n = list;
  list = list->next;
  int ret = recurse(list, prepend(dest, n), func, extra);
  n->next = list;
  list = n;

  // loop for next
  for(snode *prev = list; prev->next; prev = prev->next) {

    n = prev->next;
    prev->next = prev->next->next;

    ret = recurse(list, prepend(dest, n), func, extra);

    n->next = prev->next;
    prev->next = n;

    if (!ret)
      return 0;
  }

  return 1;
}


int permute(snode *list, snode_predicate func, void *extra)
{
  return recurse(list, 0, func, extra);
}


void destroy(snode *list)
{
  snode *temp;
  while (list) {
    temp = list;
    list = list->next;
    delete temp;
  }
}

