// SLIST.HPP v1.1
// - Scott McPeak, June 1995
// List operations for singly-linked lists
// It should be noted that there are several special-purpose list functions
// which could be derived from other functions (e.g. deleteThat); the reason
// for this is that the derived version is usually worst-case O(n) but the
// special version is structured to run in (n) time and (1) space

#ifndef __SLIST_HPP
#define __SLIST_HPP


// basic list node
class snode {
public:
  snode *next;
  static int count;

  snode() { next=0; count++; }
  virtual ~snode();
};


// data types
typedef int (*snode_predicate)(snode*, void*);
typedef snode* (*snode_map)(snode*, void*);
typedef int (*snode_differencer)(snode *left, snode *right, void*);

// selectors
int length(snode *list);               // if list is NULL, returns 0
snode *last(snode *list);
snode *nth(snode *list, int which);
int listok(snode *list);               // no loops, basically

// functions
snode *firstThat(snode *list, snode_predicate func, void *extra=0);
  // applies func to each, return first nonzero or NULL if none are

void forAll(snode *list, snode_predicate func, void *extra=0);
  // applies func to each, until func returns zero or end of list is reached

snode *map(snode *list, snode_map func, void *extra=0);
  // returns a new list made up of nodes returned by func
  // this can be used to copy a list if used with a map function that copies
  // also something of a filter; if func returns zero it won't be used

int permute(snode *list, snode_predicate func, void *extra=0);
  // permutes the list, calling func for each
  // if func returns 0, it stops
  // note factorial behavior (10 items, 10! calls to func)

// mutators (they return a ptr that is the new list after modification)
snode *prepend(snode *list, snode *toPrepend);
snode *append(snode *list, snode *toAppend);
snode *remove(snode *list, snode *toRemove);    // detach but not delete
snode *insertat(snode *list, snode *toInsert, int destPos);

snode *concat(snode *list1, snode *list2);

snode *insertionSort(snode *list, snode_differencer diff, void *extra=0);

snode *deleteThat(snode *list, snode_predicate pred, void *extra=0);
snode *removeThat(snode *list, snode **removednodes,
                  snode_predicate pred, void *extra=0);

void destroy(snode *list);        // delete (free memory) all nodes


// ----------------------- template stuff -----------------------------
// typed versions, like Next, that eliminate the need for a typecast
// T must be derived from snode or wierd errors will be the result
// note that g++ can't handle several of these so unix support is off

#ifdef __BORLANDC__
template <class T> T *Last(T *list) { return (T*)last(list); }
template <class T> T *Nth(T *list, int which) { return (T*)nth(list, which); }

template <class T> T *FirstThat(T *list,
                                int (*T_pred)(T*, void*),
                                void *extra=0)
  { return (T*)firstThat(list, (snode_predicate)T_pred, extra); }

template <class T> T *Map(T *list,
                          T* (*T_map)(T*, void*),
                          void *extra=0)
  { return (T*)map(list, (snode_map)T_map, extra); }

template <class T> T *Prepend(T *list, T *toPrepend) { return (T*)prepend(list, toPrepend); }
template <class T> T *Append(T *list, T *toAppend) { return (T*)append(list, toAppend); }
template <class T> T *Remove(T *list, T *toRemove) { return (T*)remove(list, toRemove); }

template <class T> T *Concat(T *list1, T *list2) { return (T*)concat(list1, list2); }

template <class T> T *InsertionSort(T *list,
                                    int (*T_diff)(T*, T*, void*),
                                    void *extra=0)
  { return (T*)insertionSort(list, (snode_differencer)T_diff, extra); }

template <class T> T *DeleteThat(T *list,
                                 int (*T_pred)(T*, void*),
                                 void *extra=0)
  { return (T*)deleteThat(list, (snode_predicate)T_pred, extra); }

template <class T> T *RemoveThat(T *list,
                                 T **rnodes,
                                 int (*T_pred)(T*, void*),
                                 void *extra=0)
  { return (T*)removeThat(list, rnodes, (snode_predicate)T_pred, extra); }
#endif


// for people who like iterators!
template <class T>
class listiter {
  T *ptr;

public:
  listiter(T *list) { ptr=list; }
  void reset(T *list) { ptr=list; }
  operator T* () { return ptr; }
  T* operator -> () { return ptr; }
  void operator++ () { ptr = (T*)ptr->next; }
  void operator++ (int) { ptr = (T*)ptr->next; }
};


// for people who like list headers (like me!)
template <class T>
class slist {
public:
  T *top;

  slist() { top=0; }
  ~slist() { if (top) destroy(top); }

  operator T* () { return top; }
  T* head() { return top; }
  slist& operator= (T *newtop) { top=newtop; return *this; }
};


// a quick way to generate a node class with a single extra field (int, etc)
template <class T>
class datasnode : public snode {
public:
  T data;

  datasnode() { data=0; }
  datasnode(T aData) { data=aData; }

  datasnode<T> *Next() { return (datasnode<T>*)next; }
};

/* Revision History:

1.1   Added removeThat, deleteThat, RemoveThat, DeleteThat
1.0   Original

*/

#endif
