/* TREE23.CPP
  In a quest for speed, I have two alternatives for allocation.
    1. Standard C++ new and delete, using constructors.
    2. Fixed size array-based pool system.
  new and delete allow arbitrarily-sized trees but are slower;
  the pool limits total size but runs in 1/3 the time
*/

#include "tree23.hpp"
#include <assert.h>
#include <string.h>
#include <iomanip.h>

#ifndef FASTALLOC
#define FASTALLOC 0
#endif

#ifndef POOLSIZE
#define POOLSIZE 100
#endif

#if FASTALLOC
  #define NODEDELETE(ptr) ((void)(count--))
#else
  #define NODEDELETE(ptr) (delete ptr)
#endif

int tree23::count=0;
int tree23::realcount=0;
tree23 pool[POOLSIZE];


void initpool()
{
  tree23::realcount=0;
  tree23::count=0;
  memset(pool, 0, sizeof(pool));
}


// get a new node from the pool
// left and right may be null to create a new leaf, otherwise both must be
// non-null and point to nodes to establish as children
tree23 *newtree23(int data, tree23 *left, tree23 *right)
{
  // acquire new node
  tree23 *p;

  #if FASTCOUNT
    // use a static pool of POOLSIZE nodes for fast but small trees

  p = pool + tree23::realcount;
  tree23::realcount++;
  tree23::count++;

  p->numkeys = 1;
  p->key[0] = data;           // setup the new node
  p->child[0] = left;
  p->child[1] = right;
  assert(p->parent == 0);     // ensure it happened (by different mechanism)

  if (left) {                 // if children exist, setup their "parent" member
    left->parent = p;
    assert(right);
    right->parent = p;
  }

  #else
    // use regular allocation + constructor call
    // for slower but more flexible trees

  p = new tree23(data, left, right);
  assert(p);

  #endif

  return p;
}


// create a new tree, with one node and one key
tree23 *maketree(int data)
{
  return newtree23(data, 0,0);         // no children
}


// insert a key into the tree
tree23 *insert(tree23 *tree, int akey)
{
  assert(!tree->parent);               // just making sure...

  tree23 *p = tree;
  while (!p->isleaf()) {
    for(int i=0; i< p->numkeys; i++)
      if (akey < p->key[i])           // find correct child
        break;
    p = p->child[i];                  // continue process
  }
  p->add_key(akey);                   // (possibly) perform splitting

  p = tree;
  if (p->parent) {                     // did we add a new root?
    p = tree->parent;                    // then that's the new tree,
    NODEDELETE(tree);                    // and the current one is history
  }
  return p;
}


// the node is "overfull"- it has one too many keys and children, so this splits
// it into two new nodes and either:
//   a) replaces the single entry in parent with the two new entries, or
//   b) creates a new root node with the new entries as only children
void tree23::split()
{
  tree23 *left, *right;

  // key[1] is guaranteed to be the middle key
  left = newtree23(key[0], child[0], child[1]);
  right = newtree23(key[2], child[2], child[3]);

  if (parent)
    parent->add_child(this, key[1], left, right);    // trace up tree
  else {
    parent = newtree23(key[1], left, right);         // new root
    left->parent = parent;                           // threaded
    right->parent = parent;
  }
}


// we are asked to replace "toreplace" with the dual entries "L" and "R"
void tree23::add_child(tree23 *toreplace, int akey, tree23 *L, tree23 *R)
{
  L->parent = this;
  R->parent = this;

  for(int i=0; i<= numkeys; i++)           // look for child to replace
    if (child[i] == toreplace) {
      for(int j=2; j>i; j--) {             // run through keys/children to its right
        key[j] = key[j-1];                   // and shift them over 1
        child[j+1] = child[j];
      }

      key[i] /* j == i now */ = akey;      // put the new key in the new space
      child[i+1] = R;                        // R into newly opened space
      child[i] = L;                          // and L replaces "toreplace"

      break;
    }

  numkeys++;                // record new entry's existence
  if (numkeys == 3)         // bloated node?
    split();                  // split and continue

  NODEDELETE(toreplace);      // and kill the node
}


// leaf node gets a new key
void tree23::add_key(int akey)
{
  for(int i=0; i< numkeys; i++)
    if (akey < key[i])
      break;

  for(int j=2; j>i; j--)
    key[j] = key[j-1];
  key[i] = akey;

  numkeys++;
  if (numkeys > 2)
    split();
}


int tree23::levels()
{
  tree23 *t = this;
  int count=1;

  while (!t->isleaf()) {
    count++;
    t = t->child[0];
  }

  return count;
}


// from nonmember operator to member function
ostream& operator<< (ostream &os, tree23 &obj)
{
  return obj.insert(os);
}


int tree23::printstruct=0;

// print in a lisp-like notation
ostream& tree23::insert(ostream &os)
{
  char openchar='(', closechar=')';

  if (printstruct) {
    char *openstr  = "*([{([{([{([{([{([{([{([{([{([{([{";
    char *closestr = "*)]})]})]})]})]})]})]})]})]})]})]}";

    int L = levels();
    assert (L <= strlen(openstr));
    openchar = openstr[L];
    closechar = closestr[L];
  }

  os << openchar;
  for(int i=0; i<numkeys; i++) {
    if (!isleaf())
      os << *child[i] << " ";          // print child tree

    if (!printstruct) {
      os << key[i];                    // and key
      if ((i < numkeys-1) || !isleaf())
        os << " ";                     // and usually a space to separate
    }
    else
      if (isleaf())
        os << "*";         // just stars at leaves when printing structure
  }

  if (!isleaf())
    os << *child[i] << closechar;      // last child and close-paren
  else
    os << closechar;                   // close-paren after last key

  return os;
}


// simple indentation loop
void ind(int indent)
{
  while (indent--)
    cout << " ";
}


/* Goal:
         1
      2
         3
      4
         5
   8
         9
      10
         11
*/


// print in a sideways tree structure by in-order traversal
void tree23::treeprint(int indent)
{
  for(int i=0; i<=numkeys; i++) {
    if (!isleaf())
      child[i]->treeprint(indent+2);   // tree,

    if (i<numkeys) {
      ind(indent);
      cout << key[i] << endl;          // then child except last iteration
    }
  }
}


// verify threading (thereby catching back-edges among other things)
int tree23::check()
{
  for(int i=0; i<3; i++)
    if (child[i] && ((child[i]->parent != this) ||
                     !child[i]->check()))
      return 0;
  return 1;
}


// counts all internal nodes of the tree
int tree23::countnodes()
{
  if (isleaf())
    return 1;

  int tot=0;
  for(int i=0; i<=numkeys; i++)
    tot += child[i]->countnodes();

  return tot+1;
}


// counts all leafs of a tree
int tree23::countleaves()
{
  if (isleaf())
    return numkeys+1;

  int tot=0;
  for(int i=0; i<=numkeys; i++)
    tot += child[i]->countleaves();

  return tot;
}


void tree23::levelstats(int *arr, int depth)
{
  if (numkeys==1)
    arr[depth*2]++;
  else
    arr[depth*2 + 1]++;

  if (!isleaf())
    for (int i=0; i<=numkeys; i++)
      child[i]->levelstats(arr, depth+1);
}


#pragma argsused
void destroy(tree23 *tree)
{
#if FASTALLOC
  initpool();                          // simply zero the array

#else
  if (!tree) return;

  assert(tree->check());

  destroy(tree->child[0]);
  destroy(tree->child[1]);             // recursively delete children
  destroy(tree->child[2]);

  delete tree;                         // then kill given node
#endif
}


// ^^^^^^^^^^^^^^^^^^^^^^^ storing ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
// vvvvvvvvvvvvvvvvvvvvvvv analysis vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv

/* string notation
  If it's a single level tree, string is "(*)" or "(**)".

  First is the # of levels, (e.g. 2 or 3)
  Then, a :
  Then, count of each possible child tree index
  Then, a ,
  Then, a nonequivalency number as 0,1,2:
    If all children are of one type, this is 0
    If only two children, this is 0
    If two children are X and one is Y, then 0 = Y in middle, 1 = Y on side
    If all different (XYZ), 0= YXZ, 1= XYZ, 2= XZY
      where X, Y, and Z have increasing subtree type indices.
  Example: 3:0010020,1  Has three levels, one type 2 and two type 5s, and
                        the type 5 is on the side (i.e. 225 or 522).
*/

int notation::style()
{
//  if (s[0] == '(')
//    return s[2]==')'? NS_TWOCHILDREN : NS_THREESAME;

  if (levels() < 3)
    return NS_TWOCHILDREN;       // meaning order is unimportant

  int sum=0, concen=1;
  for(int i=2; s[i] != ','; i++) {
    sum += s[i]-'0';
    if (s[i]-'0' > concen)
      concen = s[i]-'0';
  }

  return sum==2?     NS_TWOCHILDREN :
         concen==3?  NS_THREESAME   :
         concen==2?  NS_TWOSAME     :
         /* else */  NS_NOSAME ;
}


int notation::children()
{
  int sum=0;
  for(int i=2; s[i] != ','; i++)
    sum += s[i]-'0';
  return sum;
}


int notation::childindex(int where)
{
  assert((where>=0) && (where<=2));
  assert(s[0] != '(');

  int childindex[3];
  int child=0;

  for(int i=2; s[i] != ','; i++)
    for(int j=0; j< (s[i]-'0'); j++)
      childindex[child++] = i-2;
  i++;
  int nonequiv = s[i]-'0';

  int couple, single;

  // this map is the heart of handling the nonequivalency of
  // permuting children of height 2
  int ne_map[3][3] = {
    /* nonequiv     Left child     Middle child    Right child */
      /* 0 */        1 /* Y */,     0 /* X */,      2 /* Z */,
      /* 1 */        0 /* X */,     1 /* Y */,      2 /* Z */,
      /* 2 */        0 /* X */,     2 /* Z */,      1 /* Y */
  };

  switch (style()) {
    case NS_TWOCHILDREN:
//      assert(where < 2);
      return childindex[where];

    case NS_NOSAME:
      return childindex[  ne_map [ nonequiv ]
                                 [ where /* child # */ ]];

#if 0
    // this was the bug that put us back 3-4 weeks: (!)
      switch (where) {
        case 0: return childindex[ s[i]=='0'? 1 : 0 ];
        case 1: return childindex[ s[i]-'0' ];
        case 2: return childindex[ s[i]=='2'? 1 : 0 ];
        default: return -1;
      }
#endif

    case NS_TWOSAME:
      if (childindex[0] == childindex[1])       { couple=0; single=2; }
      else if (childindex[1] == childindex[2])  { couple=1; single=0; }
      else                                      { couple=0; single=1; }

      if (s[i]=='0')        // single in middle
        return childindex[ where==1? single : couple ];
      else                  // single on right (arbitrary side)
        return childindex[ where==2? single : couple ];

    case NS_THREESAME:
      return childindex[0];         // all same

    default:
      assert(0);
      return -1;
  }
}


int notation::operator++()
{
  if (s[0] == '(') {
    if (s[2] == '*')
      return 0;
    *this = string("(**)");
    return 1;
  }

  if (nonequiv() != '0') {
    nonequiv()--;
    return 1;
  }

  int retval = 1;       // nominal continue

  for(int i=2; s[i] != ','; i++);       // right edge
  for(i--; s[i]=='0'; i--);             // rightmost nonzero digit

  if (s[i+1] == ',') {          // must flip the counter
    int carryback = s[i]-'0';
    s[i] = '0';

    for(; s[i]=='0'; i--);  // 2nd-to rightmost nonzero digit

    if (s[i] == ':') {              // no more nonzero found?
      if (carryback == 3) {
        s[i+1] = 2;            // reset the counter
        retval = 0;            // end of a ++ loop
      }
      else
        s[i+1] = '3';
    }

    else {                      // normal carryback procedure
      s[i]--;
      s[i+1] = carryback + 1 + '0';
    }
  }

  // don't have to flip, just inc
  else {
    s[i]--;
    s[i+1]++;
  }

  int st = style();
  nonequiv() = st==NS_TWOCHILDREN? '0':
               st==NS_THREESAME? '0':
               st==NS_TWOSAME? '1':
               /*st==NS_NOSAME*/ '2';

  return retval;
}


void notation::zero(int levels)
{
  if (levels == 1) {
    *this = string("(*)");
    return;
  }

  int ctypes = types(levels-1);
  setlength(ctypes + 4);

  s[0] = levels+'0';
  s[1] = ':';

  s[2] = '2';
  for(int i=1; i<ctypes; i++)
    s[i+2] = '0';

  s[i+2] = ',';
  s[i+3] = '0';        // nonequivalence starts as twochildren
}


// ------------------------ treehandle -----------------
treehandle::treehandle()
{
  index = levels = 0;
  tree = 0;
}

treehandle::~treehandle()
{
  if (tree)
    destroy(tree);
  tree=0;
}


treehandle& treehandle::set(int newindex, int newlevels)
{
  index = newindex;
  levels = newlevels;

  if (tree) { destroy(tree); tree=0; }
  if (type) type = (char*)NULL;

  return *this;
}


// notation -> index
treehandle& treehandle::set(notation newtype)
{
  set(0,0);        // free old string and tree

  if (newtype[0] == '(') {
    type = newtype;
    levels = 1;
    index = type[2]==')'? 0:1;
    return *this;
  }

  type.zero(newtype.levels());

  levels = newtype.levels();
  index = 0;

  while (type != newtype) {
    int result = type++;
    assert(result);      // should not hit the end of the possibilities
    index++;
  }

  // so now index matches newtype and type has been set properly
  return *this;
}


// tree -> notation (then -> index)
treehandle &treehandle::set(tree23 *source)
{
  levels = source->levels();

  if (levels == 1) {
    set(source->numkeys==1? 0 : 1, 1);
    return *this;
  }

  int childindex[3];
  for(int i=0; i<= source->numkeys; i++)        // get types of children
    childindex[i] = treehandle().set(source->child[i]).getindex();

  type.zero(levels);
  type.settypect(0,0);

  for(i=0; i<= source->numkeys; i++)         // record children of each type
    type[childindex[i]+2]++;            // result[i] (- {0,1,2,3} for all i

  switch (type.style()) {
    case NS_TWOSAME:
      if ((childindex[0] == childindex[1]) ||
          (childindex[1] == childindex[2]))
        type.nonequiv() = '1';
      else
        type.nonequiv() = '0';
      break;

    case NS_NOSAME:
      type.nonequiv() =
        childindex[1] < childindex[0] && childindex[1] < childindex[2] ? '0':
        childindex[1] > childindex[0] && childindex[1] > childindex[2] ? '2':
                            /* else Y is in the middle */                '1';
      break;

    default: type.nonequiv() = '0';
  }

  return set(type);       // call the string -> index function
}


tree23 *treehandle::gettree()
{
  if (tree)
//    return tree;
    destroy(tree);           // for now, assume caching is not working
  tree = 0;

  gettype();                 // make sure we've got the string representation

  tree = new tree23(0,0,0);
  if (levels == 1) {
    tree->numkeys = 1 + (type == string("(**)"));
    return tree;
  }

  for(int i=0; i< type.children(); i++) {
    treehandle t;
    tree->child[i] = t.set( type.childindex(i), levels-1 ).gettree();
    t.forgettree();
    tree->child[i]->parent = tree;
  }

  tree->numkeys = i-1;
  assert(i==3 || i==2);

  return tree;
}


notation treehandle::gettype()
{
  if (tree)
    { destroy(tree); tree=0; }

  type.zero(levels);
  for(int i=0; i<index; i++)
    assert(type++);

  return type;
}


int treehandle::operator++()
{
  if (index+1 >= types(levels))
    return 0;

  index++;
  set(index, levels);
  return 1;
}


int types(int level)
{
  switch (level) {
    case 1: return 2;                 // (*) or (**)
    case 2: return 7;                 // 2:20,0 , 3:21,1 , etc.
    case 3: return 224;               // 3:2000000,0 , 3:0101010,2 , etc.
//    case 4: return 6000000;
    default: assert(0);
             return 0;
  }
}


int intpow(int base, int exp)
{
  int result = 1;
  while (exp--)
    result *= base;
  return result;
}


// ------------------------
tree23 *source, *equivptr;
int *arr;
int currentrow;
int rows, cols;
int printtypes;

tree23 *treecopy(tree23 *source, tree23 *tofind);
void recurserecord(tree23 *t);

void printcharacteristics(treehandle &h)
{
  currentrow = 0;
//  source = makeleveltree(levels, typestring(levels, typeindex));
  source = h.gettree();
  int levels = h.getlevels();

  rows = intpow(3, levels-1);
  cols = types(levels)+1;

  arr = new int[ rows * cols ];
  for(int row=0; row<rows; row++) {
    arr[row*cols+0] = 0;             // initial probibility of 0
    for(int col=1; col<cols; col++)
      arr[row*cols+col] = (col-1 == h.getindex()? -1 : 0);
  }

  recurserecord(source);

  for(row=0; row < rows; row++)
    for(int look=row+1; look < rows; look++) {
      if (!arr[look*cols+0])
        continue;

      for(int col=1; col<cols; col++)
        if (arr[row*cols+col] != arr[look*cols+col])
          break;

      if (col == cols) {           // all were equal
        arr[row*cols+0] += arr[look*cols+0];
        arr[look*cols+0] = 0;
      }
    }

  for(row=0; row<rows; row++)
    if (arr[row*cols+0]) {
      cout << h.getindex();        // print index
      if (printtypes)
        cout << " (" << h.gettype() << ")";
      cout << endl;

      cout << setw(4) << arr[row*cols+0] << "  ";
      for(int col=1; col<cols; col++)
        cout << setw(3) << arr[row*cols+col];
      cout << endl;
    }

  delete arr;
}


void recurserecord(tree23 *t)
{
  if (!t->isleaf()) {
    for(int i=0; i<= t->numkeys; i++)
      recurserecord(t->child[i]);
    return;
  }

  // is a leaf
  arr[currentrow*cols+0] = t->numkeys+1;      // record actual probability

  tree23 *working = treecopy(source, t);      // copy the tree
  equivptr->add_key(0);                       // and fake the insertion

  if (!working->parent) {
    treehandle t;
    t.set(working);      // tree -> type -> index

    arr[currentrow*cols + 1 + t.getindex()]++;
         // deleted = levelindex(levels, leveltype(levels, working))
    destroy(working);
  }

  else {                     // added a new root as well
    treehandle h;

    h.set(working->parent->child[0]);
    arr[currentrow*cols + 1 + h.getindex()]++;
//      levelindex(levels, leveltype(levels, working->parent->child[0]))]++;

    h.set(working->parent->child[1]);
    arr[currentrow*cols + 1 + h.getindex()]++;
//      levelindex(levels, leveltype(levels, working->parent->child[1]))]++;

    destroy(working->parent);      // kill the real tree
    NODEDELETE(working);           // just a dangling node
  }

  currentrow++;
}


tree23 *treecopy(tree23 *source, tree23 *tofind)
{
  tree23 *root = new tree23(0,0,0);

  if (!source->isleaf())
    for(int i=0; i<= source->numkeys; i++) {
      root->child[i] = treecopy(source->child[i], tofind);
      root->child[i]->parent = root;
    }

  root->numkeys = source->numkeys;
  if (source == tofind)
    equivptr = root;

  return root;
}















// -------------------------------------------------------------------------
#if 0
string level1type(tree23 *tree)
{
  if (tree->numkeys == 1)
    return string("10");
  else
    return string("01");
}


string leveltype(int level, tree23 *tree)
{
  if (level == 1)
    return level1type(tree);           // base case

  int childtype[3];
  for(int i=0; i<= tree->numkeys; i++)        // get types of children
    childtype[i] = levelindex(level-1, leveltype(level-1, tree->child[i]));

  int ct = childtypes(level);
  string result(ct);
  for(i=0; i<ct; i++)                  // init result to "000....000"
    result[i] = '0';
  result[ct] = 0;

  for(i=0; i<= tree->numkeys; i++)            // record children of each type
    result[childtype[i]]++;            // result[i] (- {0,1,2} for all i

  return result;
}


int not_implemented=0;           // provides a handy string for asserting

int levelindex(int level, string type)
{
  switch (level) {
    case 1: return type == string("01");      // this is string 1 (2nd of 2)

    case 2: if ( (type[0]-'0') + (type[1]-'0') == 2 )
              return type[1]-'0';             // 2nd digit is type
            else
              return type[1]-'0'+3;           // 2nd digit is type-3

    case 3: for(int i=0; i<112; i++)
              if (typestring(3, i) == type

    default: assert(not_implemented);         // if this is executed prog
             return 0;                        // will halt and print
  }                                           // the "not_implemented" string
}


tree23 *makeleveltree(int level, string type)
{
  tree23 *root = new tree23(0,0,0);
  if (level == 1) {
    root->numkeys = 1 + (type == string("01"));
    return root;
  }

  int child=0;

  for(int i=0; i< childtypes(level); i++)
    while (type[i] >= '1') {
      root->child[child] = makeleveltree(level-1, typestring(level-1, i));
      root->child[child]->parent = root;
      child++;
      type[i]--;
    }

  root->numkeys = child-1;
  assert(child==3 || child==2);

  return root;
}


// for instance, typestring(2,5) = 12
string typestring(int level, int targetindex)
{
  if (level == 1) {
    if (targetindex)
      return string("01");
    else
      return string("10");
  }

  int ct = childtypes(level);
  int data_error=0;

  string s(ct);
  for(int i=0; i<ct; i++)
    s[i] = '0';
  s[ct] = 0;

  // initialize counter to first index
  s[0] = '2';
  int index = 0;

  while (index != targetindex) {
    int rightmost1 = ct-1;
    while (s[rightmost1] == '0')
      rightmost1--;

    if (rightmost1 == ct-1) {       // flip the counter locally
      int carryback = s[ct-1]-'0';
      s[ct-1] = '0';

      while ((rightmost1 >= 0) && (s[rightmost1] == '0'))
        rightmost1--;

      if (rightmost1 == -1) {
        if (carryback == 3)
          assert(data_error);     // this is the last one
        else
          s[0] = '3';
      }

      else {                      // normal carryback procedure
        s[rightmost1]--;
        s[rightmost1+1] = carryback + 1 + '0';
      }
    }

    // don't have to flip, just inc
    else {
      s[rightmost1]--;
      s[rightmost1+1]++;
    }

    index++;
  }

  return s;
}


int levelindex(int level, string targetstring)
{
  if (level == 1) {
    if (targetstring == string("01"))
      return 1;
    else
      return 0;
  }

  int ct = childtypes(level);
  int data_error=0;

  string s(ct);
  for(int i=0; i<ct; i++)
    s[i] = '0';
  s[ct] = 0;

  // initialize counter to first index
  s[0] = '2';
  int index = 0;

  while (s != targetstring) {
    int rightmost1 = ct-1;
    while (s[rightmost1] == '0')
      rightmost1--;

    if (rightmost1 == ct-1) {       // flip the counter locally
      int carryback = s[ct-1]-'0';
      s[ct-1] = '0';

      while ((rightmost1 >= 0) && (s[rightmost1] == '0'))
        rightmost1--;

      if (rightmost1 == -1) {
        if (carryback == 3)
          assert(data_error);     // this is the last one
        else
          s[0] = '3';
      }

      else {                      // normal carryback procedure
        s[rightmost1]--;
        s[rightmost1+1] = carryback + 1 + '0';
      }
    }

    // don't have to flip, just inc
    else {
      s[rightmost1]--;
      s[rightmost1+1]++;
    }

    index++;
  }

  return index;
}


// --------------------------
tree23 *source, *equivptr;
int *arr;
int currentrow;
int rows, cols;

void printcharacteristics(int levels, int typeindex)
{
  currentrow = 0;
  source = makeleveltree(levels, typestring(levels, typeindex));
  rows = intpow(3, levels-1);
  cols = childtypes(levels+1)+1;

  arr = new int[ rows * cols ];
  for(int row=0; row<rows; row++) {
    arr[row*cols+0] = 0;             // initial probibility of 0
    for(int col=1; col<cols; col++)
      arr[row*cols+col] = (col-1 == typeindex? -1 : 0);
  }

  recurserecord(source);

  for(row=0; row < rows; row++)
    for(int look=row+1; look < rows; look++) {
      if (!arr[look*cols+0])
        continue;

      for(int col=1; col<cols; col++)
        if (arr[row*cols+col] != arr[look*cols+col])
          break;

      if (col == cols) {           // all were equal
        arr[row*cols+0] += arr[look*cols+0];
        arr[look*cols+0] = 0;
      }
    }

  for(row=0; row<rows; row++)
    if (arr[row*cols+0]) {
      cout << setw(4) << arr[row*cols+0] << "  ";
      for(int col=1; col<cols; col++)
        cout << setw(3) << arr[row*cols+col];
      cout << endl;
    }

  delete arr;
}
#endif


/* Analysis
  Ok, a bit on type notation. Types are represented as strings (why? easier
to debug). The first digit is the number of children of type 0, the next is
# of type 1, and up to the total number of distinct child tree types.
  For 1-level trees, we have either (*) or (**): "10" or "01", where type
0 is (*) and type 1 is (**).
  For 2-level, there are seven ways to arrange either 2 or 3 children, but
the number of child types is still 2- (*) or (**). So we have things like
20 = ((*) (*)), 02 = ((**) (**)), 12 = ((**) (*) (**)), etc [order irrelevnt]
  For 3-levels, a total of 112 distinct trees exist, building on the 7
possible child trees, with types like 1001000 = (1 4), 0200010 = (1 5 1),
etc.
  The idea throughout is that we only want to keep track of how many of each
type of child we have, but not their order in the child list since that is
irrelevant.
*/
