// TREE23.HPP
// 2-3 tree data structures
// purpose is to analyze the structure itself, rather than actually make
// use of the tree for storing data

#ifndef __TREE23_HPP
#define __TREE23_HPP

#include <iostream.h>
#include "string.hpp"

class tree23 {
public:
  int numkeys;               // normally 1 or 2; 3 is a possible intermediate
  int key[3];                // keys, one extra for temp
  tree23 *child[4];          // child nodes, one extra for temp
  tree23 *parent;            // threaded tree pointer

  static int count;          // how many currently in use
  static int realcount;      // count for array-based memory management

  tree23(int data, tree23 *ch0, tree23 *ch1)
    { numkeys=1;
      key[0]=data; key[1]=key[2]=0;
      child[0]=ch0; child[1]=ch1; child[2]=child[3]=0;
      if (ch0) ch0->parent = this;
      if (ch1) ch1->parent = this;
      parent=0;
      count++; }
  tree23() {}       // no count here because this is for pool
  ~tree23() { count--; }

  int isleaf() { return !child[0]; }
    // leaf when viewed without the real data keys; really just the
    // second-to-last row in a tree that stores information

  int levels();
    // counts the number of levels in the tree (single node = 1 level)

  // --- the next three functions perform splitting, an essentially
  //     linear process (recursion used just to simplify the code) ---

//private:       // these would not be used by a program that stored values
  void split();
    // this function splits the node in half, calls add_child

  void add_child(tree23 *toreplace, int key, tree23 *L, tree23 *R);
    // replaces one child with two and calls split if it needs to

public:
  void add_key(int key);
    // begins the splitting process

  static int printstruct;          // print {[(*)(**)][(*)(*)]}, not (()(..)
  friend ostream& operator<< (ostream &os, tree23 &obj);
  ostream& insert(ostream &os);
    // wrangle printing into a member function

  void treeprint(int indent);
    // print in a formatted style so structure is evident

  int check();                         // verify threading
  int countnodes();                    // count total nodes
  int countleaves();
  void levelstats(int *arr, int depth=0);
};


// --- tree api ---
tree23 *maketree(int data);
  // return a new tree with the single key 'data'

tree23 *insert(tree23 *tree, int key);
  // make a toplevel insertion; seeks to the tree's bottom and calls add_key

void destroy(tree23 *tree);
  // destroy a tree (may call initpool)


int types(int levels);                 // distinct tree types with X levels

// the essential tool for the analysis is the "notation" object, which allows
// the program to manipulate higer-level descriptions of trees
class notation : public string {
public:
  int levels() { return s[0]=='('? 1 : s[0]-'0'; }
  int childindex(int which);      // 0=left, 1=mid/right, 2=right

  int style();    // returns one of
    #define NS_TWOCHILDREN 11
    #define NS_THREESAME   3        // numbers chosen to be
    #define NS_TWOSAME     2        // helpful in debugger
    #define NS_NOSAME      0
  int children();

  char& nonequiv() { return s[ types(levels()-1)+3 ]; }

  int typect(int index) { return s[index+2] - '0'; }
  void settypect(int index, int newct) { s[index+2] = newct+'0'; }

  int operator++ ();       // meant for while(obj++) loops
  void zero(int levels);   // reset as a counter

  notation& operator= (char*src)
    { string::operator=(src); return *this; }
};


class treehandle {
  int index;
  int levels;
  tree23 *tree;
  notation type;

public:
  treehandle();
  ~treehandle();

  treehandle& set(int newindex, int newlevels);  // store the number
  treehandle& set(notation newtype);             // string -> number, store
  treehandle& set(tree23 *source);               // tree -> string -> index, store

  int getindex() { return index; }
  int getlevels() { return levels; }
  tree23 *gettree();                   // index -> string -> tree
  notation gettype();                  // index -> string
  void forgettree() { tree=0; }

  int operator++ ();                   // intended for while(obj++) loops
};


// ---- analysis functions ----
int types(int levels);                 // distinct tree types with X levels
int intpow(int base, int exponent);    // simplistic exponentiator

void printcharacteristics(treehandle &h);
  // probablistic analysis of each possible insertion into the tree


#if 0      // obsolete, absorbed by the treehandle class
string level1type(tree23 *tree);
string leveltype(int level, tree23 *tree);
int levelindex(int level, string type);
string typestring(int level, int targetindex);
tree23 *makeleveltree(int level, string type);
tree23 *treecopy(tree23 *source, tree23 *tofind);
#endif


// --- internal use ---
tree23 *newtree23(int data, tree23 *l, tree23 *r);
  // create a new tree with given data

void initpool();
  // zero the pool

extern tree23 pool[100];

#endif
